import os
import cv2
import time
import pandas as pd
import numpy as np
from astro_ghost.PS1QueryFunctions import *
from astro_ghost.NEDQueryFunctions import getNEDInfo
from astropy import units as u
from astropy.io import fits, ascii
from astropy.utils.exceptions import AstropyUserWarning,AstropyWarning
from astropy.wcs import WCS
from astropy.utils.data import get_pkg_data_filename
from matplotlib import pyplot as plt
from matplotlib import colors
from astropy.stats import sigma_clipped_stats, SigmaClip
from photutils.detection import DAOStarFinder
from photutils.background import MeanBackground, MedianBackground, Background2D
from photutils.aperture import CircularAperture
from joblib import dump, load
import gc
import warnings

def updateStep(px, gradx, grady, step, point, size):
    """ Determine direction of movement by image gradients.
    :param px: The maximum size of the image, in pixels.
    :type px: int
    :param gradx: The horizontal gradient of the image.
    :type gradx: float
    :param grady: The vertical gradient of the image.
    :type grady: float
    :param step: The step size for updating the position.
    :type step: float
    :param point: The current position.
    :type point: array-like
    :param size: The predicted size of the true host. Can be "small", "medium", or "large".
    :type size: str
    :return: The updated position.
    :rtype: array-like
    """

    max_x = px
    max_y =  px
    grad = np.array([gradx[point[0], point[1]], grady[point[0], point[1]]])
    ds = step/np.sqrt(grad[0]**2 + grad[1]**2)
    ds = np.nanmin([ds, step])
    newPoint = [point[0] + ds*grad[0], point[1] + ds*grad[1]]
    newPoint = [int(newPoint[0]), int(newPoint[1])] #round to nearest index
    if (newPoint[0] >= max_x) or (newPoint[1] >= max_y) or (newPoint[0] < 0) or (newPoint[1] < 0):
        #if we're going to go out of bounds, don't move
        return point
    #if we're stuck, perturb one pixel in a random direction:
    elif ((newPoint == point) and (size == 'large')):
        a = np.random.choice([-1, 0, 1], 2)#
        newPoint = [newPoint[0] + a[0], newPoint[1] + a[1]]
    return newPoint


def dist(p1, p2):
    """Measure the euclidean distance between two points.

    :param p1: The first point.
    :type p1: array-like
    :param p2: The second point.
    :type p2: array-like
    :return: The euclidean distance.
    :rtype: float
    """

    return np.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)

def plot_DLR_vectors_GD(size, path, transient, transient_df, host_dict_candidates, host_dict_final, host_df, R_dict, ra_dict, scale=1):
    """Plots the DLR vectors associated with each candidate host in a transient's field.

    :param size: Size of the plotted image, in pixels.
    :type size: int
    :param path: Filepath where image will be saved.
    :type path: str
    :param transient: Name of transient.
    :type transient: str
    :param transient_df: Dataframe of transient properties from TNS.
    :type transient_df:  Pandas DataFrame
    :param host_dict_candidates: key,value pairs of transient name, list of PS1 objIDs of
        candidate hosts. Dictionary is before gradient ascent is run.
    :type host_dict_candidates: dictionary
    :param host_dict_final: key,value pairs of transient name, list of PS1 objIDs of
        candidate hosts. Dictionary is after gradient ascent is run.
    :type host_dict_final: dictionary
    :param host_df: List of all candidate hosts in the field (before gradient ascent).
    :type host_df: Pandas DataFrame
    :param R_dict: The r/DLR normalized distance metrics for all candidate hosts.
    :type R_dict: dictionary
    :param ra_dict: The DLR values for all candidate hosts.
    :type ra_dict: dictionary
    :param scale: An additional scale factor for the image size (to fully capture low-z hosts).
    :type scale: float
    :return: figure axis
    :rtype: axes of the plotted image (to overlay a quiver map, or 2D field of gradient arrows).
    """

    hostList = host_dict_candidates[str(transient)]
    if type(hostList) is np.ndarray:
        if len(hostList) > 1:
            chosen = host_dict_final[transient]
        else:
            chosen = hostList[0]
    else:
        chosen = hostList
        hostList = np.array(hostList)
    band = 'r'
    px = int(size*scale)
    row = transient_df[transient_df['Name'] == transient]

    tempRA = Angle(row.RA, unit=u.degree)
    tempDEC = Angle(row.DEC, unit=u.degree)
    transientRA = tempRA.degree[0]
    transientDEC = tempDEC.degree[0]
    searchRA = transientRA
    searchDEC = transientDEC

    a = find_all("PS1_ra={}_dec={}_{}arcsec_{}.fits".format(searchRA, searchDEC, int(px*0.25), band), path)
    if not a:
        get_PS1_Pic(path, 0, searchRA, searchDEC, px, band)
        a = find_all("PS1_ra={}_dec={}_{}arcsec_{}.fits".format(searchRA, searchDEC, int(px*0.25), band), path)
    hdu = fits.open(a[0])[0]
    image_file = get_pkg_data_filename(a[0])
    image_data = fits.getdata(image_file, ext=0)
    wcs = WCS(hdu.header)
    fig = plt.figure(num=None, figsize=(20,20), facecolor='w', edgecolor='k')
    axes_coords = [0, 0, 1, 1] # plotting full width and height
    ax = fig.add_axes(axes_coords, projection=wcs)
    axes_coords2 = [-0.045, -0.03, 1.06, 1.08]
    ax_grads = fig.add_axes(axes_coords2, projection=None)
    plt.axis('off')
    for host in hostList:
        hostDF = host_df[host_df['objID'] == host]
        band = choose_band_SNR(hostDF)
        XX = hostDF[band + 'momentXX'].values[0]
        YY = hostDF[band + 'momentYY'].values[0]
        XY = hostDF[band + 'momentXY'].values[0]
        U = np.float(XY)
        Q = np.float(XX) - np.float(YY)
        if (Q == 0):
            r_a = 1.e-5
        else:
            phi = 0.5*np.arctan(U/Q)
        kappa = Q**2 + U**2
        a_over_b = (1 + kappa + 2*np.sqrt(kappa))/(1 - kappa)

        r_a = ra_dict[host]
        r_b = r_a/(a_over_b)

        hostDF['r_a'] = r_a
        hostDF['r_b'] = r_b
        hostDF['phi'] = phi
        hostRA = host_df.loc[host_df['objID'] == host,'raMean'].values[0]
        hostDEC = host_df.loc[host_df['objID'] == host,'decMean'].values[0]

        hostDLR = R_dict[host]
        c = '#666dc9'
        c2 = 'red'
        if (host == chosen):
            c = c2 = '#d308d0'
        hostDF['raMean'], hostDF['decMean']
        plot_ellipse(ax, px, hostDF, searchRA, searchDEC, c)

        # in arcseconds
        dx = float(hostRA - transientRA)*3600
        dy = float(hostDEC - transientDEC)*3600

        dist = np.sqrt(dx**2 + dy**2)
        if hostDLR == 10000000000.0:
            hostDLR = 0.0
        else:
            hostDLR = dist/hostDLR

        #in arcseconds
        scale_factor = hostDLR/dist

        DLR_RA = float(hostRA) - dx*scale_factor/3600
        DLR_DEC = float(hostDEC) - dy*scale_factor/3600

        pointRA = [hostRA, DLR_RA]
        pointDEC = [hostDEC, DLR_DEC]

        ax.plot(pointRA, pointDEC, transform=ax.get_transform('fk5'), lw=6, color= c)
    ax.imshow(image_data, norm=colors.PowerNorm(gamma = 0.5, vmin=1, vmax=1.e4), cmap='gray')
    plt.axis('off')
    return ax_grads


def plot_ellipse(ax, px, s, ra, dec, color):
    """Plots an ellipse on an image.

    :param ax: Axis of figure for plotting.
    :type ax: figure axis
    :param px: Image size, in pixels.
    :type px: int
    :param s: PS1 source, with shape parameters (phi, r_a, and r_b)
    :type s: Pandas DataFrame
    :param ra: Right ascension of image center, in degrees.
    :type ra: float
    :param dec: Declination of image center, in degrees.
    :type dec: float
    :param color: Color of plotted ellipse.
    :type color: str
    """

    i=0
    size = px  #PS cutout image size, 240*sidelength in arcmin
    x0, y0 = ((ra-s['raMean'])*4*3600*np.cos(s['decMean']/180*np.pi)+(size/2)), (s['decMean']-dec)*4*3600+(size/2)
    i=i+1

    y, x = np.mgrid[0:size, 0:size]# 4 pixel for 1 arcsec for PS1, here image size is set to be 20"*20", depend on your cutout image size
    #make fitted image
    n_radius=2
    theta1 = s['phi']#rot angle
    a1= s['r_a']
    b1= s['r_b']
    e1 = mpatches.Ellipse((x0, y0), 4*n_radius*a1, 4*n_radius*b1, theta1, lw=6, ls='--', edgecolor=color,
                          facecolor='none',  label='source 1')
    ax.add_patch(e1)


def denoise(img, weight=0.1, eps=1e-3, num_iter_max=200):
    """Perform total-variation denoising on a grayscale image.
       Uses Rudin, Osher and Fatemi algorithm.

    :param img: 2-D input data to be de-noised.
    :type img: array-like
    :param weight: Denoising weight. The greater `weight`, the more de-noising (at
        the expense of fidelity to `img`).
    :type weight: float, optional
    :param eps: Relative difference of the value of the cost function that determines
        the stop criterion. The algorithm stops when:
        (E_(n-1) - E_n) < eps * E_0
    :type eps: float, optional
    :param num_iter_max: Maximal number of iterations used for the optimization.
    :type num_iter_max: int, optional
    :return: De-noised array of floats.
    :rtype: array-like
    """

    u = np.zeros_like(img)
    px = np.zeros_like(img)
    py = np.zeros_like(img)

    nm = np.prod(img.shape[:2])
    tau = 0.125

    i = 0
    while i < num_iter_max:
        u_old = u

        # x and y components of u's gradient
        ux = np.roll(u, -1, axis=1) - u
        uy = np.roll(u, -1, axis=0) - u

        # update the dual variable
        px_new = px + (tau / weight) * ux
        py_new = py + (tau / weight) * uy
        norm_new = np.maximum(1, np.sqrt(px_new **2 + py_new ** 2))
        px = px_new / norm_new
        py = py_new / norm_new

        # calculate divergence
        rx = np.roll(px, 1, axis=1)
        ry = np.roll(py, 1, axis=0)
        div_p = (px - rx) + (py - ry)

        # update image
        u = img + weight * div_p

        # calculate error
        error = np.linalg.norm(u - u_old) / np.sqrt(nm)

        if i == 0:
            err_init = error
            err_prev = error
        else:
            # break if error small enough
            if np.abs(err_prev - error) < eps * err_init:
                break
            else:
                e_prev = error

        # don't forget to update iterator
        i += 1

    return u

def get_clean_img(path, ra, dec, px, band):
    """Takes PS1 images, removes bad pixels, and
       estimates new pixel values through a 2D
       interpolation.

    :param path: filepath where the image will be saved.
    :type path: str
    :param ra: Right ascension of image position, in degrees.
    :type ra: float
    :param dec: Declination of image position, in degrees.
    :type dec: float
    :param px: Image size, in pixels.
    :type px: int
    :param band: Passband of image.
    :type band: str
    :return: Image data, with bad pixels masked.
    :rtype: 2D array
    :return: Astropy world coordinate system object
    :rtype: wcs of image fits file.
    :return: Header of image fits file.
    :rtype: Fits header
    """

    #first, mask the data
    if dec > -30:
        a = find_all("PS1_ra={}_dec={}_{}arcsec_{}.fits".format(ra, dec, int(px*0.25), band), path)
        if not a:
            get_PS1_Pic(path, 0, ra, dec, px, band)
            a = find_all("PS1_ra={}_dec={}_{}arcsec_{}.fits".format(ra, dec, int(px*0.25), band), path)
        b_fn = "PS1_ra={}_dec={}_{}arcsec_{}_stack.mask.fits".format(ra, dec, int(px*0.25), band)
        b = find_all(b_fn, path)
        if not b:
            get_PS1_type(path, ra, dec, px, band, 'stack.mask')
            b = find_all(b_fn, path)
        c_fn = "PS1_ra={}_dec={}_{}arcsec_{}_stack.num.fits".format(ra, dec, int(px*0.25), band)
        c = find_all(c_fn, path)
        if not c:
            get_PS1_type(path, ra, dec, px, band, 'stack.num')
            c = find_all(c_fn, path)

        image_data_mask = fits.open(b[0])[0].data
        image_data_num = fits.open(c[0])[0].data
        image_data = fits.open(a[0])[0].data

        hdu = fits.open(a[0])[0]
        wcs = WCS(hdu.header)

        bit = image_data_mask
        mask = image_data_mask
        for i in np.arange(np.shape(bit)[0]):
            for j in np.arange(np.shape(bit)[1]):
                if image_data_mask[i][j] == image_data_mask[i][j]:
                    bit[i][j] = "{0:016b}".format(int(image_data_mask[i][j]))
                    tempBit = str(bit[i][j])[:-2]
                    if len(str(int(bit[i][j]))) > 12:
                        if (tempBit[-6] == 1) or (tempBit[-13] == 1):
                            mask[i][j] = np.nan
                    elif len(str(int(bit[i][j]))) > 5:
                        if (tempBit[-6] == 1):
                            mask[i][j] = np.nan

        mask = ~np.isnan(image_data_mask)
        mask_num = image_data_num
        image_masked = np.ma.masked_array(image_data, mask=mask)
        image_masked_num = np.ma.masked_array(image_masked, mask=mask_num)

    else:
        print("Error! Gradient Ascent is not yet implemented for southern-hemisphere sources.")
        a = find_all("SkyMapper_ra={}_dec={}_{}arcsec_{}.fits".format(ra, dec, int(px*0.5), band), '.')
        if not a:
            get_SkyMapper_Pic(0, ra, dec, px, band)
            a = find_all("SkyMapper_ra={}_dec={}_{}arcsec_{}.fits".format(ra, dec, int(px*0.5), band), '.')
        b = find_all("SkyMapper_ra={}_dec={}_{}arcsec_{}_mask.fits".format(ra, dec, int(px*0.5), band), '.')
        if not b:
            get_SkyMapper_mask(ra, dec, px, band)
            b = find_all("SkyMapper_ra={}_dec={}_{}arcsec_{}_mask.fits".format(ra, dec, int(px*0.5), band), '.')
        image_data_mask = fits.open(b[0])[0].data
        image_data = fits.open(a[0])[0].data

        hdu = fits.open(a[0])[0]
        wcs = WCS(hdu.header)

        bit = image_data_mask
        mask = image_data_mask
        for i in np.arange(np.shape(bit)[0]):
            for j in np.arange(np.shape(bit)[1]):
                if image_data_mask[i][j] == image_data_mask[i][j]:
                    bit[i][j] = "{0:016b}".format(int(image_data_mask[i][j]))
                    tempBit = str(bit[i][j])[:-2]
                    if len(str(int(bit[i][j]))) > 12:
                        if (tempBit[-6] == 1) or (tempBit[-13] == 1):
                            mask[i][j] = np.nan
                    elif len(str(int(bit[i][j]))) > 5:
                        if (tempBit[-6] == 1):
                            mask[i][j] = np.nan
        mask = ~np.isnan(image_data_mask)
        image_masked = np.ma.masked_array(image_data, mask=mask)

    #then return the data
    return np.array(image_masked), wcs, hdu

def getSteps(SN_dict, transientNames, hostDF):
    """Calculates a scale factor for the gradient ascent step
       based on the mean kron radius of the galaxies in the field.

    :param SN_dict: Key,value pairs of transient names and lists of candidate
        host galaxy objIDs in PS1.
    :type SN_dict: dictionary
    :param transientNames: Names of transients to associate.
    :type transientNames: array-like
    :param hostDF: Dataframe of all candidate host galaxies.
    :type hostDF: Pandas DataFrame
    :return: List of calculated steps.
    :rtype: array-like
    """

    steps = []
    hostDF.replace(-999, np.nan, inplace=True)
    hostDF.replace(-999, np.nan, inplace=True)
    for name in transientNames:
        hostList = SN_dict[name]
        if (type(hostList) is np.int64 or type(hostList) is float):
            hostList = [hostList]
        checkNan = [x == x for x in hostList]
        if np.sum(checkNan) > 0:
            hostRadii = hostDF.loc[hostDF['objID'].isin(hostList), 'rKronRad'].values
            mean = np.nanmean(hostRadii)
            if mean == mean:
                mean = np.max([mean,2])
                step = np.min([mean, 50])

                #find some proper scaling factor between the mean and the step size
                steps.append(step)
            else:
                steps.append(5)
        else:
            steps.append(5)
    return steps

def gradientAscent(path, SN_dict, SN_dict_postDLR, transientNames, hostDF, transientDF, fn, plot=True, px=800):
    """The gradient ascent algorithm for identifying a final host galaxy.

    :param path: Filepath to save the output log for the algorithm.
    :type path: str
    :param SN_dict: Key, val pairs are transient name, list of PS1 objIDs of
        potential hosts. Dictionary is before the DLR method.
    :type SN_dict: dictionary
    :param SN_dict_postDLR: Key, val pairs are transient name, list of PS1 objIDs of
        potential hosts. Dictionary is after the DLR method (so
        should have mostly one host per transient).
    :type SN_dict_postDLR: dictionary
    :param transientNames: List of transient names.
    :type transientNames: array-like
    :param hostDF: PS1 info for candidate hosts.
    :type hostDF: Pandas DataFrame
    :param transientDF: TNS info for associated transients.
    :type transientDF: Pandas DataFrame
    :param fn: Filename for gradientAscent log.
    :type fn: str
    :param plot: If True, plot the associated transients, the background maps,
        and the gradient fields associated with each transient.
    :type plot: bool, optional
    :param px: Image size (in pixels).
    :type px: int
    :return: Dictionary of transient, host objID pairs after
        gradient ascent.
    :rtype: dictionary
    :return: Dataframe of final associated hosts (after gradient ascent).
    :rtype: Pandas Dataframe
    :return: List of transients for which gradient ascent failed.
    :rtype: array-like
    """

    warnings.filterwarnings('ignore', category=AstropyUserWarning)
    warnings.filterwarnings('ignore', category=AstropyWarning)
    step_sizes = getSteps(SN_dict, transientNames, hostDF)
    unchanged = []
    N_associated = 0
    totalDTime = 0.
    totalGATime = 0.
    startGATime = time.time()
    f = open(fn, 'w')
    print("Starting size of data frame: %i" % len(hostDF), file=f)
    if plot:
        try:
            os.makedirs('quiverMaps')
        except:
            print("Already have the folder quiverMaps!")
    for i in np.arange(len(step_sizes)):
        #clear buffer memory
        gc.collect()

        transient_name = transientNames[i]
        print("Transient: %s"% transient_name, file=f)

        ra = transientDF.loc[transientDF['Name'] == transient_name, 'RA'].values[0]
        dec = transientDF.loc[transientDF['Name'] == transient_name, 'DEC'].values[0]

        startTime = time.time()
        g_img, wcs, g_hdu  = get_clean_img(path, ra, dec, px, 'g')
        g_mask = np.ma.masked_invalid(g_img).mask
        r_img, wcs, r_hdu  = get_clean_img(path, ra, dec, px, 'r')
        r_mask = np.ma.masked_invalid(r_img).mask
        i_img, wcs, i_hdu  = get_clean_img(path, ra, dec, px, 'i')
        i_mask = np.ma.masked_invalid(i_img).mask
        endTime = time.time()
        dTime = endTime - startTime
        print("Download time: %.2f\n"%dTime, file=f)
        totalDTime+=dTime

        # cleanup - remove the fits files when we're done using them
        for band in ['g', 'r', 'i']:
            os.remove(path + "/PS1_ra={}_dec={}_{}arcsec_{}_stack.num.fits".format(ra, dec, int(px*0.25), band))
            os.remove(path + "/PS1_ra={}_dec={}_{}arcsec_{}_stack.mask.fits".format(ra, dec, int(px*0.25), band))
            os.remove(path + "/PS1_ra={}_dec={}_{}arcsec_{}.fits".format(ra, dec, int(px*0.25), band))

        nancount = 0
        obj_interp = []
        print("Commencing DAOStarFinder:\n", file=f)
        for obj in [g_img, r_img, i_img]:
            data = obj
            mean, median, std = sigma_clipped_stats(data, sigma=15.0)
            daofind = DAOStarFinder(fwhm=3.0, threshold=15.*std)
            sources = daofind(data - median)
            try:
                xvals = np.array(sources['xcentroid'])
                yvals = np.array(sources['ycentroid'])
                for k in np.arange(len(xvals)):
                    tempx = xvals[k]
                    tempy = yvals[k]
                    yleft = np.max([int(tempy) - 10, 0])
                    yright = np.min([int(tempy) + 10, np.shape(data)[1]-1])
                    xleft = np.max([int(tempx) - 10, 0])
                    xright = np.min([int(tempx) + 10, np.shape(data)[1]-1])

                    for r in np.arange(yleft,yright+1):
                        for j in np.arange(xleft, xright+1):
                            if dist([xvals[k], yvals[k]], [j, r]) < 7:
                                data[r, j] = np.nan
                nancount += np.sum(np.isnan(data))
                positions = np.transpose((sources['xcentroid'], sources['ycentroid']))
                apertures = CircularAperture(positions, r=7.)
                if plot:
                    fig = plt.figure(figsize=(10,10))
                    ax = fig.gca()
                    ax.imshow(data)
                    apertures.plot(color='blue', lw=1.5, alpha=0.5)
                    plt.axis('off')
                    plt.savefig("quiverMaps/detectedStars_%s.png"%transient_name, bbox_inches='tight')
                    plt.close()
            except:
                print("No stars here!", file=f)

            # new way of interpolating the data
            mask = np.isnan(data).astype(np.uint8)
            data = cv2.inpaint(data.astype(np.float32), mask, 3, cv2.INPAINT_TELEA)

            obj_interp.append(data)

        gMax = np.nanmax(obj_interp[0])

        g_ZP = g_hdu.header['ZPT_0001']
        r_ZP = r_hdu.header['ZPT_0001']
        i_ZP = i_hdu.header['ZPT_0001']

        gmag = -2.5*np.log10(obj_interp[0]) + g_ZP
        rmag = -2.5*np.log10(obj_interp[1]) + r_ZP
        imag = -2.5*np.log10(obj_interp[2]) + i_ZP

        #now the mean can be taken
        mean_zp = (g_ZP + r_ZP + i_ZP)/3
        meanMag = (gmag + rmag + imag)/3
        meanImg = 10**((mean_zp-meanMag)/2.5) #convert back to flux

        print("NanCount = %i"%nancount,file=f)
        mean_center = meanImg[int(px/2),int(px/2)]
        print("Mean_center = %f" % mean_center,file=f)

        if plot:
            plt.figure(figsize=(10,7))
            plt.hist(meanImg.flatten(), bins=np.logspace(0, 6))
            plt.xscale("log")
            plt.savefig("quiverMaps/countsHistogram_%s.png" % transient_name, bbox_inches='tight')
            plt.close()

        meanImg[meanImg != meanImg] = 1.e-30
        mean, median, std = sigma_clipped_stats(meanImg, sigma=10.0)
        print("mean image = %e"% mean, file=f)
        aboveCount = np.sum(meanImg > 1.)
        aboveCount2 = np.sum(meanImg[int(px/2)-100:int(px/2)+100, int(px/2)-100:int(px/2)+100] > 1.)
        aboveFrac2= aboveCount2/40000
        print("aboveCount = %f"% aboveCount,file=f)
        print("aboveCount2 = %f "% aboveCount2, file=f)
        totalPx = px**2
        aboveFrac = aboveCount/totalPx
        print("aboveFrac= %f" % aboveFrac, file=f)
        print("aboveFrac2 = %f "% aboveFrac2, file=f)

        if ((median <15) and (np.round(aboveFrac2, 2) < 0.70)) or ((mean_center > 1.e3) and (np.round(aboveFrac,2) < 0.60) and (np.round(aboveFrac2,2) < 0.75)):
            bs = 10
            fs = 1
            if aboveFrac2 < 0.7:
                step_sizes[int(i)] = 2.
            else:
                step_sizes[int(i)] = 10.
            print("Small filter", file=f)
            size = 'small'
        elif ((mean_center > 40) and (median > 500) and (aboveFrac > 0.60)) or ((mean_center > 300) and (aboveFrac2 > 0.7)):
            bs = 75 #the big sources
            fs = 3
            print("Large filter", file=f)
            step_sizes[int(i)] = np.max([step_sizes[int(i)], 50])
            size = 'large'
        else:
            bs = 40 #everything in between
            fs = 3
            print("Medium filter", file=f)
            step_sizes[int(i)] = np.max([step_sizes[int(i)], 15])
            size = 'medium'

        sigma_clip = SigmaClip(sigma=15.)
        bkg_estimator = MeanBackground()
        bkg3_g = Background2D(g_img, box_size=bs, filter_size=fs,
         sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)
        bkg3_r = Background2D(r_img, box_size=bs, filter_size=fs,
         sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)
        bkg3_i = Background2D(i_img, box_size=bs, filter_size=fs,
         sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)

        #Bkg is in counts so average in mags
        bkg3_g.background[bkg3_g.background < 0] = 1.e-30
        bkg3_r.background[bkg3_r.background < 0] = 1.e-30
        bkg3_i.background[bkg3_i.background < 0] = 1.e-30

        backmag_g = -2.5*np.log10(bkg3_g.background) + g_ZP
        backmag_r = -2.5*np.log10(bkg3_r.background) + r_ZP
        backmag_i = -2.5*np.log10(bkg3_i.background) + i_ZP

        mean_zp = (g_ZP + r_ZP + i_ZP)/3.
        backmag = 0.333*backmag_g + 0.333*backmag_r + 0.333*backmag_i
        background = 10**(mean_zp-backmag/2.5)

        if plot:
            fig, axs = plt.subplots(1, 3, sharex=True, sharey=True,figsize=(20,10))
            axs[0].imshow(bkg3_g.background)
            axs[0].axis('off')
            axs[1].imshow(bkg3_r.background)
            axs[1].axis('off')
            axs[2].imshow(bkg3_i.background)
            axs[2].axis('off')
            plt.savefig("quiverMaps/backgrounds_%s.png" % transient_name, bbox_inches='tight')
            plt.close()

        mean, median, std = sigma_clipped_stats(meanImg, sigma=1.0)
        meanImg[meanImg <= (mean)] = 1.e-30
        meanImg[meanImg < 0] = 1.e-30

        if plot:
            fig = plt.figure(figsize=(10,10))
            ax = fig.gca()
            ax.imshow((meanImg)/np.nanmax(meanImg))
            plt.axis('off')
            plt.savefig("quiverMaps/normalizedMeanImage_%s.png" % transient_name, bbox_inches='tight')
            plt.close()

            fig = plt.figure(figsize=(10,10))
            ax = fig.gca()
            ax.imshow(background/np.nanmax(background))
            plt.axis('off')
            plt.savefig("quiverMaps/normalizedMeanBackground_%s.png" % transient_name, bbox_inches='tight')
            plt.close()

        if nancount > 1.e5:
            imgWeight = 0
        elif (mean_center > 1.e4):
            imgWeight = 0.75
        elif size == 'medium':
            imgWeight = 0.33
        else:
            imgWeight = 0.10

        print("imgWeight= %f"%imgWeight, file=f)
        fullbackground = ((1-imgWeight)*background/np.nanmax(background) + imgWeight*meanImg/np.nanmax(meanImg))*np.nanmax(background)

        n = px
        X, Y = np.mgrid[0:n, 0:n]
        dx, dy = np.gradient(fullbackground.T)

        n_plot = 10

        dx_small = dx[::n_plot, ::n_plot]
        dy_small = dy[::n_plot, ::n_plot]
        print("step = %f"%  step_sizes[int(i)], file=f)

        # the center of the grid
        start = [[int(px/2),int(px/2)]]

        if True:
            start.append(updateStep(px, dx, dy, step_sizes[int(i)], start[-1], size))
            for j in np.arange(1.e3):
                start.append(updateStep(px, dx, dy, step_sizes[int(i)], start[-1], size))
            it_array = np.array(start)
            endPoint = start[-1]

            if plot:
                fig  = plt.figure(figsize=(10,10))
                ax = fig.gca()
                ax.imshow(fullbackground)
                plt.axis("off")
                plt.savefig("quiverMaps/fullBackground_%s.png"%transient_name, bbox_inches='tight')
                plt.close()

            coords = wcs.wcs_pix2world(endPoint[0], endPoint[1], 0., ra_dec_order = True) # Note the third argument, set to 0, which indicates whether the pixel coordinates should be treated as starting from (1, 1) (as FITS files do) or from (0, 0)
            print("Final ra, dec after GD : %f %f"% (coords[0], coords[1]), file=f)

            col = '#D34E24'
            col2 = '#B54A24'

            #lookup by ra, dec
            try:
                if size == 'large':
                    a = ps1cone(float(coords[0]), float(coords[1]), 20/3600)
                else:
                    a = ps1cone(float(coords[0]), float(coords[1]), 5/3600)
            except TypeError:
                 continue
            if a:
                print("Found a host here!", file=f)
                a = ascii.read(a)
                a = a.to_pandas()
                a.sort_values(by=['distance'], inplace=True)

                a = a[a['nDetections'] > 1]
                smallType = ['AbLS', 'EmLS' , 'EmObj', 'G', 'GammaS', 'GClstr', 'GGroup', 'GPair', 'GTrpl', 'G_Lens', 'IrS', 'PofG', 'RadioS', 'UvES', 'UvS', 'XrayS', '', 'QSO', 'QGroup', 'Q_Lens']
                medType = ['G', 'IrS', 'PofG', 'RadioS', 'GPair', 'GGroup', 'GClstr', 'EmLS', 'RadioS', 'UvS', 'UvES', '']
                largeType = ['G', 'PofG', 'GPair', 'GGroup', 'GClstr']
                if len(a) > 0:
                    a = getNEDInfo(a)
                    if (size == 'large'):
                        print("L: picking the closest NED galaxy within 20 arcsec", file=f)

                        tempA = a[a['NED_type'].isin(largeType)]
                        if len(tempA) > 0:
                            a = tempA
                        tempA = a[a['NED_type'] == 'G']
                        if len(tempA) > 0:
                            a = tempA

                        if len(a) > 1:
                            a = a.iloc[[0]]
                    elif (size == 'medium'):
                        print("M: Picking the closest NED galaxy within 5 arcsec", file=f)
                        tempA = a[a['NED_type'].isin(medType)]
                        if len(tempA) > 0:
                            a = tempA
                        if len(a) > 1:
                            a = a.iloc[[0]]
                    else:
                        tempA = a[a['NED_type'].isin(smallType)]
                        if len(tempA) > 0:
                            a = tempA
                        a = a.iloc[[0]]
                        print("S: Picking the closest non-stellar source within 5 arcsec", file=f)
                    print("Nice! Host association chosen.", file=f)
                    print("NED type: %s" % a['NED_type'].values[0], file=f)
                    print(a['objID'].values[0], file=f)
                    print("Chosen Host RA and DEC: %f %f"% (a['raMean'], a['decMean']), file=f)
                    SN_dict_postDLR[transient_name] = a['objID'].values[0]
                    print("Dict value: %i"%SN_dict_postDLR[transient_name],file=f)
                    N = len(hostDF)
                    hostDF = pd.concat([hostDF, a], ignore_index=True)
                    N2 = len(hostDF)
                    if N2 != (N+1):
                        print("ERROR! Value not concatenated!!", file=f)
                        return
                    finalRA = np.array(a['raMean'])
                    finalDEC = np.array(a['decMean'])
                    col = 'tab:green'
                    col2 = '#078840'
                else:
                    unchanged.append(transient_name)
            else:
                unchanged.append(transient_name)
            if plot:
                fig = plt.figure(figsize=(20,20))
                ax = fig.gca()
                ax.imshow(i_img, norm=colors.PowerNorm(gamma = 0.5, vmin=1, vmax=1.e4), cmap='gray')
                it_array = np.array(start)
                ax.plot(it_array.T[0], it_array.T[1], "--", lw=5, c=col, zorder=20)
                ax.scatter([int(px/2)], [int(px/2)], marker='*', s=1000, color='#f3a712', zorder=50)
                ax.scatter(endPoint[0], endPoint[1], marker='*', lw=4, s=1000, facecolor='#f3a712', edgecolor=col2, zorder=200)
                ax.quiver(X[::n_plot,::n_plot], Y[::n_plot,::n_plot], dx[::n_plot,::n_plot], dy[::n_plot,::n_plot], color='#845C9B', angles='xy', scale_units = 'xy')
                plt.axis('off')
                plt.savefig("quiverMaps/quiverMap_%s.png"%transient_name, bbox_inches='tight')
                plt.close()
            N_associated += 1
        f.flush()
        if N_associated%10 == 0:
            print("N_associated = %i"%N_associated,file=f)
            print("Size of table = %i"%len(hostDF),file=f)
            print("Saving the modified way!")
            with open(path+"/dictionaries/gals_postGD.p", 'wb') as fp:
                dump(SN_dict_postDLR, fp)
            hostDF.to_csv(path+"/hostDF_postGD.tar.gz", index=False)

    with open(path+"/dictionaries/gals_postGD.p", 'wb') as fp:
        dump(SN_dict_postDLR, fp)
    hostDF.to_csv(path+"/hostDF_postGD.tar.gz", index=False)
    print("Total Download time: %.2f\n"%totalDTime, file=f)
    endGATime = time.time()
    totalGATime = endGATime - startGATime
    print("Total Gradient Ascent time:%.2f\n"%totalGATime, file=f)
    f.close()
    return  SN_dict_postDLR, hostDF, unchanged



from astropy.table import Table
#import pkg_resources
import sys
import re
import numpy as np
import pylab
import json
import requests
from astropy.io import fits
from astro_ghost.PS1QueryFunctions import *

try: # Python 3.x
    from urllib.parse import quote as urlencode
    from urllib.request import urlretrieve
    import http.client as httplib
except ImportError:  # Python 2.x
    from urllib import pathname2url as urlencode
    from urllib import urlretrieve
    import httplib

import pandas as pd
import tensorflow as tf
from tensorflow import keras

from concurrent.futures import ThreadPoolExecutor
import asyncio
import codecs

import time
import sfdmap
import os
import tarfile

def build_sfd_dir(fname='./sfddata-master.tar.gz'):
    """Downloads directory of Galactic dust maps for extinction correction.

    :param fname: Filename for dustmaps.
    :type fname: str
    """

    url = 'https://github.com/kbarbary/sfddata/archive/master.tar.gz'
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        with open(fname, 'wb') as f:
            f.write(response.raw.read())
    tar = tarfile.open(fname)
    tar.extractall()
    tar.close()
    os.remove(fname)
    print("Done creating dust directory.")
    return

def get_photoz_weights(fname='./MLP_lupton.hdf5'):
    """Get weights for MLP photo-z model.

    :param fname: Filename of saved MLP weights.
    :type fname: str
    """

    url = 'https://uofi.box.com/shared/static/n1yiy818mv5b5riy2h3dg5yk2by3swos.hdf5'
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        with open(fname, 'wb') as f:
            f.write(response.raw.read())
    print("Done getting photo-z weights.")
    return

def ps1objIDsearch(objID,table="mean",release="dr1",format="csv",columns=None,
           baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs", verbose=False,
           **kw):
    """Do an object lookup by objID.
   
    :param objID: list of objIDs (or dictionary?)
    :type objID: List of objIDs
    :param table: Can be \\'mean\\', \\'stack\\', or \\'detection\\'.
    :type table: str
    :param release: Can be 'dr1' or 'dr2'.
    :type release: str
    :param format: Can be 'csv', 'votable', or 'json'
    :type format: str
    :param columns: list of column names to include (None means use defaults)
    :type columns: arrray-like
    :param baseurl: base URL for the request
    :type baseurl: str
    :param verbose: print info about request
    :type verbose: bool,optional
    :param \\*\\*kw: other parameters (e.g., 'nDetections.min':2)
    :type \\*\\*kw: dictionary
    """

    #this is a dictionary... we want a list of dictionaries
    objID=list(objID)

    data_list=[kw.copy() for i in range(len(objID))]
    assert len(data_list)==len(objID)

    for i in range(len(data_list)):
        data_list[i]['objID'] = objID[i]

    urls = []
    datas = []
    for i in range(len(objID)):
        url, data = ps1search(table=table,release=release,format=format,columns=columns,
                    baseurl=baseurl, verbose=verbose, **data_list[i])

        urls.append(url)
        datas.append(data)

    return urls, datas

def fetch_information_serially(url, data, verbose=False, format='csv'):
    """A helper function called by serial_objID_search-- Queries PanStarrs API for data.
    
    :param url: Remote PS1 url.
    :type url: str
    :param data: List of objIDs requesting
    :type data: list
    :param verbose: If True,
    :type verbose: bool, optional
    :param format: Can be \\'csv\\', \\'json\\', or \\'votable\\'.
    :type format: str
    :return:
    :rtype: str in format given by \\'format\\'.
    """

    results = []
    for i in range(len(url)):
        r = requests.get(url[i], params=data[i])
        if verbose:
            print(r.url)
        r.raise_for_status()
        if format == "json":
            results.append(r.json())
        else:
            results.append(r.text)

    return results

def post_url_parallel(results,YSE_ID):
    """TODO: unused function. The querying of PS servers is the slowest part of the script. How to parallelize?
     
    :param results: Description of parameter.
    :type results: type
    :param YSE_ID: Description of parameter.
    :type YSE_ID: type
    :return: Description of returned object.
    :rtype: Pandas DataFrame
    """

    if type(results) != str:
        results = codecs.decode(results,'UTF-8')
    lines = results.split('\n')
    print(lines)
    if len(lines) > 2:
        values = [line.strip().split(',') for line in lines]
        DF = pd.DataFrame(values[1:-1],columns=values[0])
    else:
        print('No Matches')
        DF = pd.DataFrame()
    DF['YSE_id'] = np.ones(len(DF))*YSE_ID
    return DF

def post_url_serial(results,YSE_ID):
    """A helper function called by serial_objID_search. Post-processes the data retrieved from PS1 Servers into a pandas.DataFrame object.

    :param results: The string resulting from PS1 query.
    :type results: str
    :param YSE_ID: local integer used for as an index tracking user objects vs retrived objects.
    :type YSE_ID: int
    :return: DataFrame object of the retrieved data from PS1 servers
    :rtype: pandas.DataFrame
    """
    if type(results) != str:
        results = codecs.decode(results,'UTF-8')
    lines = results.split('\r\n')
    #print(lines)
    if len(lines) > 2:
        values = [line.strip().split(',') for line in lines]
        DF = pd.DataFrame(values[1:-1],columns=values[0])
    else:
        DF = pd.DataFrame()
    DF['id'] = np.ones(len(DF))*YSE_ID
    return DF

def serial_objID_search(objIDs,table='forced_mean',release='dr2',columns=None,verbose=False,**constraints):
    """Given a list of ObjIDs, queries the PS1 server these object's Forced Mean Photometry, then returns matches as a pandas.DataFrame.
    
    :param objIDs: list of PS1 objIDs for objects user would like to query
    :type objIDs: list
    :param table: Which table to perform the query on. Default 'forced_mean'
    :type table: str
    :param release: Which release to perform the query on. Default 'dr2'
    :type release: str
    :param columns: list of what data fields to include; None means use default columns. Default None
    :type columns: list or None
    :param verbose: boolean setting level of feedback user received. default False
    :type verbose: bool
    :param \\*\\*constraints: Keyword dictionary with an additional constraints for the PS1 query
    :type \\*\\*constraints: dict
    :return: list of pd.DataFrame objects. If a match was found, then the Dataframe contains data, else it only contains a local integer.
    :rtype: pd.DataFrame
    """

    constrains=constraints.copy()
    URLS, DATAS = ps1objIDsearch(objID=objIDs,table='forced_mean',release=release,columns=columns,verbose=verbose,**constraints)
    Return = fetch_information_serially(URLS,DATAS)
    DFs=[]
    for i in range(len(Return)):
        DFs.append(post_url_serial(Return[i],i))

    return DFs

def post_url_parallel(results,YSE_ID):
    """TODO: unused function. The querying of PS servers is the slowest part of the script. How to parallelize?

    :param results: Description of param.
    :type results: type
    :param YSE_ID: Description of param.
    :type YSE_ID: type
    :return:
    :rtype:
    """

    results = codecs.decode(results,'UTF-8')
    lines = results.split('\n')
    if len(lines) > 2:
        values = [line.strip().split(',') for line in lines]
        DF = pd.DataFrame(values[1:-1],columns=values[0])
    else:
        DF = pd.DataFrame()
    DF['id'] = np.ones(len(DF))*YSE_ID
    return DF

def get_common_constraints_columns():
    """Helper function that returns a dictionary of constraints used for the matching objects in PS1 archive, and the columns of data we requre.

    :return: dictionary with our constaint that we must have more than one detection
    :rtype: dict
    :return: List of PS1 fields required for matching and NN inputs
    :rtype: list
    """

    constraints = {'nDetections.gt':1}

    #objects with n_detection=1 sometimes just an artifact.
    # strip blanks and weed out blank and commented-out values
    columns ="""objID, raMean, decMean, gFKronFlux, rFKronFlux, iFKronFlux, zFKronFlux, yFKronFlux,
    gFPSFFlux, rFPSFFlux, iFPSFFlux, zFPSFFlux, yFPSFFlux,
    gFApFlux, rFApFlux, iFApFlux, zFApFlux, yFApFlux,
    gFmeanflxR5, rFmeanflxR5, iFmeanflxR5, zFmeanflxR5, yFmeanflxR5,
    gFmeanflxR6, rFmeanflxR6, iFmeanflxR6, zFmeanflxR6, yFmeanflxR6,
    gFmeanflxR7, rFmeanflxR7, iFmeanflxR7, zFmeanflxR7, yFmeanflxR7""".split(',')
    columns = [x.strip() for x in columns]
    columns = [x for x in columns if x and not x.startswith('#')]

    return constraints, columns

def preprocess(DF,PATH='../DATA/sfddata-master/', ebv=True):
    """Preprocesses the data inside pandas.DataFrame object returned by serial_objID_search to the space of Inputs of our Neural Network.

    :param DF: Dataframe object containing the data for each matched objID
    :type DF: pandas DataFrame
    :param PATH: string path to extinction maps data
    :type PATH: str
    :param ebv: boolean for lookup of extinction data. If False, all extinctions set to 0.
    :type ebv: False
    :return: Preprocessed inputs ready to be used as input to NN
    :rtype: numpy ndarray
    """
    if ebv:
        m = sfdmap.SFDMap(PATH)
        assert ('raMean' in DF.columns.values) and ('decMean' in DF.columns.values), 'DustMap query failed because the expected coordinates didnt'\
                                                                            'exist in DF, likely the match of any Hosts into PanStarrs failed'
        EBV = m.ebv(DF['raMean'].values.astype(np.float32),DF['decMean'].values.astype(np.float32))

        DF['ebv'] = EBV
    else:
        DF['ebv'] = 0.0

    def convert_flux_to_luptitude(f,b,f_0=3631):
        return -2.5/np.log(10) * (np.arcsinh((f/f_0)/(2*b)) + np.log(b))

    b_g = 1.7058474723241624e-09
    b_r = 4.65521985283191e-09
    b_i = 1.2132217745483221e-08
    b_z = 2.013446972858555e-08
    b_y = 5.0575501316874416e-08


    MEANS = np.array([18.70654578, 17.77948707, 17.34226094, 17.1227873 , 16.92087669,
           19.73947441, 18.89279411, 18.4077393 , 18.1311733 , 17.64741402,
           19.01595669, 18.16447837, 17.73199409, 17.50486095, 17.20389615,
           19.07834251, 18.16996592, 17.71492073, 17.44861273, 17.15508793,
           18.79100201, 17.89569908, 17.45774026, 17.20338482, 16.93640741,
           18.62759241, 17.7453392 , 17.31341498, 17.06194499, 16.79030564,
            0.02543223])

    STDS = np.array([1.7657395 , 1.24853534, 1.08151972, 1.03490545, 0.87252421,
           1.32486758, 0.9222839 , 0.73701807, 0.65002723, 0.41779001,
           1.51554956, 1.05734494, 0.89939638, 0.82754093, 0.63381611,
           1.48411417, 1.05425943, 0.89979008, 0.83934385, 0.64990996,
           1.54735158, 1.10985163, 0.96460099, 0.90685922, 0.74507053,
           1.57813401, 1.14290345, 1.00162105, 0.94634726, 0.80124359,
           0.01687839])

    data_columns = ['gFKronFlux', 'rFKronFlux', 'iFKronFlux', 'zFKronFlux', 'yFKronFlux',
    'gFPSFFlux', 'rFPSFFlux', 'iFPSFFlux', 'zFPSFFlux', 'yFPSFFlux',
    'gFApFlux', 'rFApFlux', 'iFApFlux', 'zFApFlux', 'yFApFlux',
    'gFmeanflxR5', 'rFmeanflxR5', 'iFmeanflxR5', 'zFmeanflxR5', 'yFmeanflxR5',
    'gFmeanflxR6', 'rFmeanflxR6', 'iFmeanflxR6', 'zFmeanflxR6', 'yFmeanflxR6',
    'gFmeanflxR7', 'rFmeanflxR7', 'iFmeanflxR7', 'zFmeanflxR7', 'yFmeanflxR7', 'ebv']

    X = DF[data_columns].values.astype(np.float32)
    X[:,0:30:5] = convert_flux_to_luptitude(X[:,0:30:5],b=b_g)
    X[:,1:30:5] = convert_flux_to_luptitude(X[:,1:30:5],b=b_r)
    X[:,2:30:5] = convert_flux_to_luptitude(X[:,2:30:5],b=b_i)
    X[:,3:30:5] = convert_flux_to_luptitude(X[:,3:30:5],b=b_z)
    X[:,4:30:5] = convert_flux_to_luptitude(X[:,4:30:5],b=b_y)

    X = (X-MEANS)/STDS
    X[X>20] = 20
    X[X<-20] = -20
    X[np.isnan(X)] = -20

    return X

def load_lupton_model(model_path):
    """Helper function that defines and loads the weights of our NN model and the output space of the NN.

    :param model_path: path to the model weights.
    :type model_path: str
    :return: Trained photo-z MLP.
    :rtype: tensorflow keras Model
    :return: Array of binned redshift space corresponding to the output space of the NN
    :rtype: numpy ndarray
    """

    build_sfd_dir()
    get_photoz_weights(model_path)
    def model():
        INPUT = tf.keras.layers.Input(31)

        DENSE1 = tf.keras.layers.Dense(256,activation=tf.keras.layers.LeakyReLU(),kernel_initializer=tf.keras.initializers.he_normal(),kernel_regularizer=tf.keras.regularizers.l2(1e-5))(INPUT)
        DROP1 = tf.keras.layers.Dropout(0.05)(DENSE1)

        DENSE2 = tf.keras.layers.Dense(1024,activation=tf.keras.layers.LeakyReLU(),kernel_initializer=tf.keras.initializers.he_normal(),kernel_regularizer=tf.keras.regularizers.l2(1e-5))(DROP1)
        DROP2 = tf.keras.layers.Dropout(0.05)(DENSE2)

        DENSE3 = tf.keras.layers.Dense(1024,activation=tf.keras.layers.LeakyReLU(),kernel_initializer=tf.keras.initializers.he_normal(),kernel_regularizer=tf.keras.regularizers.l2(1e-5))(DROP2)
        DROP3 = tf.keras.layers.Dropout(0.05)(DENSE3)

        DENSE4 = tf.keras.layers.Dense(1024,activation=tf.keras.layers.LeakyReLU(),kernel_initializer=tf.keras.initializers.he_normal(),kernel_regularizer=tf.keras.regularizers.l2(1e-5))(DROP3)

        OUTPUT = tf.keras.layers.Dense(360,activation=tf.keras.activations.softmax)(DENSE4)

        model = tf.keras.Model(INPUT,OUTPUT)

        return model
    mymodel = model()
    mymodel.load_weights(model_path)

    NB_BINS = 360
    ZMIN = 0.0
    ZMAX = 1.0
    BIN_SIZE = (ZMAX - ZMIN) / NB_BINS
    range_z = np.linspace(ZMIN, ZMAX, NB_BINS + 1)[:NB_BINS]

    return mymodel, range_z

def evaluate(X,mymodel,range_z):
    """Evaluate the MLP for a set of PS1 inputs, and return predictions.

    :param X: PS1 properties of associated hosts.
    :type X: array-like
    :param mymodel: MLP model for photo-z estimation.
    :type mymodel: tensorflow keras Model
    :param range_z: Grid over which to evaluate the posterior distribution of photo-zs.
    :type range_z: array-like

    :return: Posterior distributions for the grid of redshifts defined as
        \\`np.linspace(0, 1, n)\\`
    :rtype: numpy ndarray shape of (df.shape[0], n)
    :return: Means
    :rtype: numpy ndarray shape of (df.shape[0],)
    :return: Standard deviations
    :rtype: numpy ndarray shape of (df.shape[0],)
    """

    posteriors = mymodel(X,training=False).numpy()
    point_estimates = np.sum(posteriors*range_z,axis=1)
    for i in range(len(posteriors)):
        posteriors[i,:] /= np.sum(posteriors[i,:])
    errors=np.ones(len(posteriors))
    for i in range(len(posteriors)):
        errors[i] = (np.std(np.random.choice(a=range_z,size=1000,p=posteriors[i,:],replace=True)))

    return posteriors, point_estimates, errors


#'id' column in DF is the 0th ordered index of hosts. missing rows are therefore signalled
#    by skipped numbers in index
def calc_photoz(hosts):
    """PhotoZ beta: not tested for missing objids.
       photo-z uses a artificial neural network to estimate P(Z) in range Z = (0 - 1)
       range_z is the value of z
       posterior is an estimate PDF of the probability of z
       point estimate uses the mean to find a single value estimate
       error is an array that uses sampling from the posterior to estimate a std dev.
       Relies upon the sfdmap package, (which is compatible with both unix and windows),
       found at https://github.com/kbarbary/sfdmap.

    :param hosts: The matched hosts from GHOST.
    :type hosts: pandas DataFrame
    :return: The matched hosts from GHOST, with photo-z point estimates and uncertainties.
    :rtype: pandas DataFrame
    """

    if np.nansum(hosts['decMean'] < -30) > 0:
        print("ERROR! Photo-z estimator has not yet been implemented for southern-hemisphere sources."\
        "Please remove sources below dec=-30d and try again.")
        return hosts
    objIDs = hosts['objID'].values.tolist()
    constraints, columns = get_common_constraints_columns()
    DFs = serial_objID_search(objIDs,columns=columns,**constraints)
    DF = pd.concat(DFs)

    posteriors, point_estimates, errors = get_photoz(DF)
    successIDs = DF['objID'].values

    for i in np.arange(len(successIDs)):
        objID = int(successIDs[i])
        hosts.loc[hosts['objID']==objID, 'photo_z'] = point_estimates[i]
        hosts.loc[hosts['objID']==objID, 'photo_z_err'] = errors[i]
    return hosts


def get_photoz(df):
    """Evaluate photo-z model for Pan-STARRS forced photometry.

    :param df: Pan-STARRS forced mean photometry data, you can get it using
        \\`ps1objIDsearch\\` from this module, Pan-STARRS web-portal or via
        astroquery i.e., \\`astroquery.mast.Catalogs.query_{criteria,region}(...,
        catalog=\\'Panstarrs\\',table=\\'forced_mean\\')\\`
    :type df: pandas DataFrame
    :return: Posterior distributions for the grid of redshifts defined as
        \\`np.linspace(0, 1, n)\\`
    :rtype: numpy ndarray shape of (df.shape[0], n)
    :return: Means
    :rtype: numpy ndarray shape of (df.shape[0],)
    :return: Standard deviations
    :rtype: numpy ndarray shape of (df.shape[0],)
    """

    # The function load_lupton_model downloads the necessary dust models and
    # weights from the ghost server.
    dust_path = './sfddata-master'
    model_path = './MLP_lupton.hdf5'

    model, range_z = load_lupton_model(model_path)
    X = preprocess(df, dust_path)
    return evaluate(X, model, range_z)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import colors
from astro_ghost.sourceCleaning import clean_dict
from astropy import units as u
from astropy.coordinates import SkyCoord, Angle
from astropy.wcs import WCS
from astropy.utils.data import get_pkg_data_filename

def choose_band_SNR(host_df):
    """Gets the PS1 band (of grizy) with the highest SNR in PSF mag.
       From https://www.eso.org/~ohainaut/ccd/sn.html,
       Error on Mag  ~    1/  (S/N)
       So calculating S/N for each band as 1/PSFMagErr
       Estimate the S/N of each band and choose the bands
       with the highest S/N for the rest of our measurements.

    :param host_df: The dataframe containing the candidate host galaxy (should just be one galaxy).
    :type host_df: Pandas DataFrame
    :return: The PS1 band with the highest S/N.
    :rtype: str
    """

    bands = 'grizy'
    SNR = []
    try:
        for band in bands:
             SNR.append(float(1/host_df["%sKronMagErr"%band]))
        i = np.nanargmax(np.array(SNR))
    except:
        #if we have issues getting the band with the highest SNR, just use r-band
        i = 1
    return bands[i]

def calc_DLR(ra_SN, dec_SN, ra_host, dec_host, r_a, r_b, source, best_band):
    """Calculate the directional light radius for a given galaxy and transient pair. Calculation is adapted from
       Gupta et al., 2013 found at
       https://repository.upenn.edu/cgi/viewcontent.cgi?referer=https://www.google.com/&httpsredir=1&article=1916&context=edissertations.

    :param ra_SN: The right ascension of the SN, in degrees.
    :type ra_SN: float
    :param dec_SN: The declination of the SN, in degrees.
    :type dec_SN: float
    :param ra_host: The right ascension of the host, in degrees.
    :type ra_host: float
    :param dec_host: The declination of the host, in degrees.
    :type dec_host: float
    :param r_a: The semi-major axis of the host in arcseconds.
    :type r_a: float
    :param r_b: The semi-minor axis of the host in arcseconds.
    :type r_b: float
    :param source: The Dataframe containing the PS1 information for the candidate host galaxy.
    :type source: Pandas DataFrame
    :param best_band: The PS1 passband with the highest S/N, from which second-order moments are estimated.
    :type best_band: str
    :return: The angular separation between galaxy and transient, in arcseconds.
    :rtype: float
    :return: The normalized distance (angular separation divided by the DLR).
    :rtype: float
    """

    xr = (ra_SN.deg - float(ra_host))*3600
    yr = (dec_SN.deg - float(dec_host))*3600

    SNcoord = SkyCoord(ra_SN, dec_SN, frame='icrs')
    hostCoord = SkyCoord(ra_host*u.deg, dec_host*u.deg, frame='icrs')
    sep = SNcoord.separation(hostCoord)
    dist = sep.arcsecond
    badR = 10000000000.0

    XX = best_band + 'momentXX'
    YY = best_band + 'momentYY'
    XY = best_band + 'momentXY'

    # if we don't have spatial information, get rid of it
    # this gets rid of lots of artifacts without radius information
    if (float(source[XX]) != float(source[XX])) | (float(source[XY]) != float(source[XY])) | \
        (float(source[YY]) != float(source[YY])):
        return dist, badR

    U = float(source[XY])
    Q = float(source[XX]) - float(source[YY])
    if Q == 0:
        return dist, badR

    phi = 0.5*np.arctan(U/Q)
    kappa = Q**2 + U**2
    a_over_b = (1 + kappa + 2*np.sqrt(kappa))/(1 - kappa)

    gam = np.arctan2(xr, yr)

    theta = phi - gam

    DLR = r_a/np.sqrt(((a_over_b)*np.sin(theta))**2 + (np.cos(theta))**2)

    R = float(dist/DLR)

    if (R != R):
        return dist, badR

    return dist, R


def calc_DLR_SM(ra_SN, dec_SN, ra_host, dec_host, r_a, elong, phi, source, best_band):
    """ Calculate the DLR method but for Skymapper (southern-hemisphere) sources,
        which don't have xx and yy moments reported in the catalog.

    :param ra_SN: The right ascension of the SN, in degrees.
    :type ra_SN: float
    :param dec_SN: The declination of the SN, in degrees.
    :type dec_SN: float
    :param ra_host: The right ascension of the host, in degrees.
    :type ra_host: float
    :param dec_host: The declination of the host, in degrees.
    :type dec_host: float
    :param r_a: The semi-major axis of the host in arcseconds.
    :type r_a: float
    :param elong: The elongation parameter of the galaxy.
    :type elong: float
    :param phi: The rotation angle of the galaxy, in radians.
    :type phi: float
    :param source: The Dataframe containing the PS1 information for the candidate host galaxy.
    :type source: Pandas DataFrame
    :param best_band: The PS1 passband with the highest S/N, from which second-order moments are estimated.
    :type best_band: str
    :return: The angular separation between galaxy and transient, in arcseconds.
    :rtype: float
    :return: The normalized distance (angular separation divided by the DLR).
    :rtype: float
    """

    # EVERYTHING IS IN ARCSECONDS
    ## taken from "Understanding Type Ia Supernovae Through Their Host Galaxies..." by Gupta
    #https://repository.upenn.edu/cgi/viewcontent.cgi?referer=https://www.google.com/&httpsredir=1&article=1916&context=edissertations
    xr = (ra_SN.deg - float(ra_host))*3600
    yr = (dec_SN.deg - float(dec_host))*3600

    SNcoord = SkyCoord(ra_SN, dec_SN, frame='icrs')
    hostCoord = SkyCoord(ra_host*u.deg, dec_host*u.deg, frame='icrs')
    sep = SNcoord.separation(hostCoord)
    dist = sep.arcsecond
    badR = 10000000000.0

    # if we don't have shape information for a southern-hemisphere galaxy, return
    if (float(r_a) != float(r_a)) | (float(elong) != float(elong)):
        return dist, badR

    gam = np.arctan2(xr, yr)

    theta = phi - gam

    #elong == a/b, which allows us to substitute here
    DLR = r_a/np.sqrt(((elong)*np.sin(theta))**2 + (np.cos(theta))**2)

    R = float(dist/DLR)

    if (R != R):
        return dist, badR

    return dist, R

def chooseByDLR(path, hosts, transients, fn, orig_dict, todo="s"):
    """The wrapper function for selecting hosts by the directional light radius method
       introduced in Gupta et al., 2013.

    :param path: Filepath where to write out the results of the DLR algorithm.
    :type path: str
    :param hosts: DataFrame containing PS1 information for all candidate hosts.
    :type hosts: Pandas DataFrame
    :param transients: DataFrame containing TNS information for all transients.
    :type transients: Pandas DataFrame
    :param fn: Filename to write the results of the associations (useful for debugging).
    :type fn: str
    :param orig_dict: Dictionary consisting of key,val pairs of transient names, and lists of
        their candidate host galaxy objIDs in PS1.
    :type orig_dict: dictionary
    :param todo: If todo == \\'s\\', save the dictionary and the list of remaining sources.
        If todo == \\'r\\', return them.
    :type todo: str
    :return: The dataframe of PS1 properties for host galaxies found by DLR.
    :rtype: Pandas DataFrame
    :return: Dictionary of matches after DLR, with transient names as keys and a list of host galaxy pan-starrs objIDs as values.
    :rtype: dictionary
    :return: List of transients for which no reliable host galaxy was found.
    :rtype: array-like
    :return: List of transients for which an issue arose in DLR (most likely, a candidate
        host galaxy in the field didn't have radius information). This list is used
        to recommend candidates to associate via the gradient ascent method.
    :rtype: array-like
    """

    dict_mod = orig_dict.copy()
    if todo=="s":
        if not os.path.exists(path+'/dictionaries'):
             os.makedirs(path+'/dictionaries')
        if not os.path.exists(path+'/tables'):
             os.makedirs(path+'/tables')
    hosts['dist/DLR'] = np.nan
    hosts['dist'] = np.nan
    noHosts = []

    clean_dict(orig_dict, hosts, [])
    f = open(path+fn, 'w')
    GDflag = 0
    GA_SN = []
    for name, host in orig_dict.items():
        if (type(host) is not np.int64 and type(host) is not float):
             if (len(host.shape) > 0) and (host.shape[0] > 0):
                R_dict = {}
                ra_dict = {}
                dist_dict = {}
                host = np.array(host)
                if len(host)>0:
                    for tempHost in host:
                        theta = 0
                        host_df = hosts[hosts["objID"] == tempHost]
                        transient_df = transients[transients["Name"] == str(name)]

                        # If no good, we want to artificially inflate the distance to the SN so that
                        # we don't incorrectly pick this as our host
                        # (but the radius goes into plotting, so ra is artificially shrunk)
                        R =  dist = 1.e10
                        r_a = 0.05

                        #flag for running gradient ascent, if the DLR method fails.
                        GDflag = 1

                        if ":" in str(transient_df["RA"].values):
                            ra_SN = Angle(transient_df["RA"].values, unit=u.hourangle)
                        else:
                            ra_SN = Angle(transient_df["RA"].values, unit=u.deg)
                        dec_SN = Angle(transient_df["DEC"].values, unit=u.degree)
                        ra_host = host_df['raMean']
                        dec_host = host_df['decMean']
                        if len(np.array(ra_SN)) > 1:
                            ra_SN = ra_SN[0]
                            dec_SN = dec_SN[0]
                        if (dec_SN.deg > -30):
                            band = choose_band_SNR(host_df)
                            r_a = r_b = float(host_df[band + 'KronRad'].values[0])
                            dist, R = calc_DLR(ra_SN, dec_SN, ra_host, dec_host, r_a, r_b, host_df, band)
                        else:
                            band = choose_band_SNR(host_df)
                            r_a = float(host_df['%sKronRad'%band].values[0])
                            if r_a == r_a:
                                elong = host_df[band + "_elong"].values[0]
                                phi = np.radians(host_df[band + "_pa"].values[0])
                                r_a = host_df[band + 'KronRad'].values[0]*0.5 #plate scale

                                # in arcsec, the radius containing 90% of the galaxy light.
                                # This empirically has improved association performance
                                # for southern-hemisphere sources.
                                dist, R = calc_DLR_SM(ra_SN, dec_SN, ra_host, dec_host, r_a, elong, phi, host_df, band)

                        R_dict[tempHost] = R
                        ra_dict[tempHost] = r_a
                        dist_dict[tempHost] = dist

                        hosts.loc[hosts['objID'] == tempHost, 'dist/DLR'] = R
                        hosts.loc[hosts['objID'] == tempHost, 'dist'] = dist

                print("\n transient = \\", file=f)
                print(name, file=f)
                print("offset/DLR = \\", file=f)
                #round for printing purposes
                R_dict_print = {k:round(v,2) if isinstance(v,float) else v for k,v in R_dict.items()}
                print(R_dict_print, file=f)
                ra_dict_print = {k:round(v,2) if isinstance(v,float) else v for k,v in ra_dict.items()}
                print("candidate semi-major axis = \\", file=f)
                print(ra_dict_print, file=f)

                #subset so that we're less than 5 in DLR units
                chosenHost = min(R_dict, key=R_dict.get)
                if R_dict[chosenHost] > 5.0:
                    #If we can't find a host, say that this galaxy has no host
                    dict_mod[name] = np.nan
                    noHosts.append(name)
                    print("No host chosen! r/DLR > 5.0.", file=f)
                    continue
                else:
                    R_dict_sub = dict((k, v) for k, v in R_dict.items() if v <= 5.0)
                    #Sort from lowest to highest DLR value
                    R_dict_sub = {k: v for k, v in sorted(R_dict_sub.items(), key=lambda item: item[1])}

                    if len(R_dict_sub.keys()) > 1:
                        gal_hosts = []
                        Simbad_hosts = []
                        for key in R_dict_sub:
                            tempType = hosts[hosts['objID'] == key]['NED_type'].values[0]
                            hasSimbad = hosts[hosts['objID'] == key]['hasSimbad'].values[0]
                            if (tempType == "G"):
                                gal_hosts.append(key)
                            if (hasSimbad) & (tempType != '*'):
                                Simbad_hosts.append(key)
                        if len(gal_hosts) > 0:
                            if gal_hosts[0] != chosenHost and R_dict[gal_hosts[0]] < 5.0:
                                chosenHost = gal_hosts[0] #only change if we're within the light profile of the galaxy
                                print("Choosing the galaxy with the smallest DLR - nearest source had DLR > 1!", file=f)
                        if len(Simbad_hosts) > 0:
                            print("Chosen SIMBAD host!", file=f)
                            if Simbad_hosts[0] != chosenHost and R_dict[Simbad_hosts[0]] < 5.0:
                                chosenHost = Simbad_hosts[0] #only change if we're within the light profile of the galaxy
                                print("Choosing the simbad source with the smallest DLR!", file=f)
                    dict_mod[name] = chosenHost
                    hosts.loc[hosts['objID'] == chosenHost, 'dist/DLR'] = R_dict[chosenHost]
                    hosts.loc[hosts['objID'] == chosenHost, 'dist'] = dist_dict[chosenHost]
                    if (GDflag):
                        GDflag = 0
                        print("Issue with DLR. Try Gradient Descent!", file=f)
                        GA_SN.append(name)
                    print(float(hosts[hosts['objID'] == chosenHost]['raMean']), float(hosts[hosts['objID'] == chosenHost]['decMean']), file=f)
                f.flush()
    f.close()
    if todo == "s":
        with open('../dictionaries/DLR_rankOrdered_hosts.p', 'wb') as fp:
            pickle.dump(dict_mod, fp, protocol=pickle.HIGHEST_PROTOCOL)
        hosts.to_csv("../tables/DLR_rankOrdered_hosts.csv")
        return
    elif todo =="r":
        return hosts, dict_mod, noHosts, GA_SN



import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from astropy.table import Table
import scipy.interpolate as scinterp
import importlib_resources

def calc_7DCD(df):
    """Calculates the color distance (7DCD) of objects in df from the
       stellar locus from Tonry et al., 2012.

    :param df: Dataframe of PS1 objects.
    :type df: Pandas DataFrame

    :return: The same dataframe as input, with new column 7DCD.
    :rtype: Pandas DataFrame
    """
    df.replace(999.00, np.nan)
    df.replace(-999.00, np.nan)

    #read the stellar locus table from PS1
    stream = importlib_resources.files(__name__).joinpath('tonry_ps1_locus.txt')
    skt = Table.read(stream, format='ascii')

    gr = scinterp.interp1d(skt['ri'], skt['gr'], kind='cubic', fill_value='extrapolate')
    iz = scinterp.interp1d(skt['ri'], skt['iz'], kind='cubic', fill_value='extrapolate')
    zy = scinterp.interp1d(skt['ri'], skt['zy'], kind='cubic', fill_value='extrapolate')
    ri = np.arange(-0.4, 2.01, 0.001)

    gr_new = gr(ri)
    iz_new = iz(ri)
    zy_new = zy(ri)

    bands = ['g', 'r', 'i', 'z']

    #adding the errors in quadrature
    df["g-rErr"] =  np.sqrt(df["gApMagErr"].astype('float')**2 + df["rApMagErr"].astype('float')**2)
    df["r-iErr"] =  np.sqrt(df["rApMagErr"].astype('float')**2 + df["iApMagErr"].astype('float')**2)
    df["i-zErr"] =  np.sqrt(df["iApMagErr"].astype('float')**2 + df["zApMagErr"].astype('float')**2)
    df['z-yErr'] =  np.sqrt(df['zApMagErr'].astype('float')**2 + df['yApMagErr'].astype('float')**2)

    df["7DCD"] = np.nan
    df.reset_index(drop=True, inplace=True)
    for i in np.arange(len(df["i-z"])):
        temp_7DCD = []

        temp_7DCD_1val_gr = (df.loc[i,"g-r"] - gr_new)**2/df.loc[i, "g-rErr"]
        temp_7DCD_1val_ri = (df.loc[i,"r-i"] - ri)**2 /df.loc[i, "r-iErr"]
        temp_7DCD_1val_iz = (df.loc[i,"i-z"] - iz_new)**2/df.loc[i, "i-zErr"]
        temp_7DCD_1val_zy = (df.loc[i,"z-y"] - zy_new)**2/df.loc[i, "z-yErr"]

        temp_7DCD_1val = temp_7DCD_1val_gr + temp_7DCD_1val_ri + temp_7DCD_1val_iz + temp_7DCD_1val_zy

        df.loc[i,"7DCD"] = np.nanmin(np.array(temp_7DCD_1val))
    return df

def plotLocus(df, color=False, save=False, type="", timestamp=""):
    """Plots the color-color distribution of objects in df along with the Tonry et al., 2012
       PS1 stellar locus.

    :param df: Dataframe of PS1 objects.
    :type df: Pandas DataFrame
    :param color: If True, color objects by their distance from the stellar locus.
    :type color: bool, optional
    :param save: If True, saves the image.
    :type save: bool, optional
    :param type: Can be "Gals" for galaxies or "Stars" for stars. Only relevant for
        coloring the distributions.
    :type type: str, optional
    :param timestamp: Timestamp to append to the saved plot filename.
    :type timestamp: str, optional
    """
    if color:
        plt.figure(figsize=(8,8))
        plt.scatter(df["i-z"], df["g-r"], c=df["7DCD"], s=2, alpha=0.8, norm=LogNorm())
        plt.xlim(-0.75, 1)
        plt.clim(0.1, 1000)
        plt.ylim(-0.5, 2)
        plt.xlabel("i-z")
        plt.ylabel("g-r")
        cbar = plt.colorbar()
        cbar.set_label("4D Color Distance")
        if save:
            plt.savefig("PS1_%s_StellarLocus_%s_Colored.pdf"%(type, timestamp))
        else:
            plt.show()
    else:
        #read the stellar locus table from PS1
        stream = importlib_resources.files(__name__).joinpath('tonry_ps1_locus.txt')
        skt = Table.read(stream, format='ascii')

        gr = scinterp.interp1d(skt['ri'], skt['gr'], kind='cubic', fill_value='extrapolate')
        iz = scinterp.interp1d(skt['ri'], skt['iz'], kind='cubic', fill_value='extrapolate')
        zy = scinterp.interp1d(skt['ri'], skt['zy'], kind='cubic', fill_value='extrapolate')
        ri = np.arange(-0.4, 2.01, 0.001)

        gr_new = gr(ri)
        iz_new = iz(ri)
        if type == 'Gals':
            c = '#087e8b'
        elif type == "Stars":
            c = '#ffbf00'
        else:
            c = 'violet'
        g = sns.JointGrid(x="i-z", y="g-r", data=df, height=9, ratio=5, xlim=(-0.4, 0.8), ylim=(-0.2, 2.0), space=0)
        g = g.plot_joint(sns.kdeplot, color=c,gridsize=200)
        g.ax_joint.plot(iz_new, gr_new, '--', c='#8c837c', lw=2)
        plt.xlabel(r"$i-z$",fontsize=18)
        plt.ylabel(r"$g-r$",fontsize=18)
        g = g.plot_marginals(sns.kdeplot, color=c, shade=True)
        if save:
            g.savefig("PS1_%s_StellarLocus_%s.pdf"%(type, timestamp))



import matplotlib
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.ipac.ned import Ned
from astroquery.simbad import Simbad
import warnings
warnings.filterwarnings("ignore", category=UserWarning)
import re
import os
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from astro_ghost.PS1QueryFunctions import find_all

def getSimbadInfo(df):
    """Check if a given PS1 object is in Simbad.

    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :return: PS1 properties for hosts, with added column
        'hasSimbad'.
    :rtype: Pandas DataFrame
    """

    df.reset_index(inplace=True, drop=True)

    df['hasSimbad'] = 0

    missingCounter = 0

    for index, row in df.iterrows():
        tempRA = row['raMean']
        tempDEC = row['decMean']
        # create a sky coordinate to query
        c = SkyCoord(ra=tempRA*u.degree, dec=tempDEC*u.degree, frame='icrs')

        # execute query
        result_table = Simbad.query_region(c, radius=(0.00055555)*u.deg)
        if result_table:
            df.at[index, 'hasSimbad'] = 1
        if missingCounter > 5000:
            print("Locked out of Simbad, will have to try again later...")
            return df
    return df



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from setuptools import setup

version = "1.0.1"

VERSION_TEMPLATE = """
 Note that we need to fall back to the hard-coded version if either
 setuptools_scm can't be imported or setuptools_scm can't determine the
 version, so we catch the generic 'Exception'.
__version__ = '{version}'
""".lstrip()

#setup(
#    use_scm_version={'write_to': os.path.join('astro_ghost', 'version.py'),
#                     'write_to_template': VERSION_TEMPLATE},

#)
#__version__ = vs

import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="astro_ghost",
    version=version,
    author="Alex Gagliano",
    author_email="gaglian2@illinois.edu",
    description="A package to associate transients with host galaxies, and a database of 16k SNe-host galaxies in PS1.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    package_data={
    'astro_ghost': ['Star_Galaxy_RealisticModel_GHOST_PS1ClassLabels.sav','tonry_ps1_locus.txt'],
    },
    install_requires=['pandas', 'scikit-learn', 'numpy', 'seaborn', 'matplotlib', 'joypy','astropy', 'photutils', 'scipy', 'datetime', 'requests<2.29.0','imblearn','rfpimp','Pillow', 'pyvo', 'astroquery', 'mastcasjobs', 'opencv-python','tensorflow','sfdmap','importlib_resources'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
#    use_scm_version={'write_to': os.path.join('astro_ghost', 'version.py'),
#                     'write_to_template': VERSION_TEMPLATE},
)



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle
import os
import sys
from datetime import datetime
import pickle
from collections import Counter

def build_ML_df(dic, hostDF, transientDF):
    """Consolidates the final host associations into a single dataframe.

    :param dic: key,value pairs of transient name, list of associated host PS1 objIDs
        (should be one-to-one except where the association failed).
    :type dic: dictionary
    :param hostDF: PS1 properties for all host galaxies.
    :type hostDF: Pandas DataFrame
    :param transientDF: Pandas DataFrame
    :type transientDF: TNS properties for all transients.
    :return: The final consolidated DF of transient & host galaxy properties.
    :rtype: Pandas DataFrame
    """

    hostDF = hostDF.reset_index(drop=True)
    hostDF = hostDF.drop_duplicates(subset=['objID'],ignore_index=True)
    hostDF["TransientClass"] = ""
    hostDF["TransientName"] = ""
    colNames = set(transientDF.columns.values)
    colNames.remove('HostName')
    colNames.remove('RA')
    colNames.remove('DEC')
    colNames.remove('Obj. Type')
    for name, host in dic.items():
        # only do matching if there's a found host
        chosenHost = ""
        if (host == host):
            if isinstance(host, np.ndarray):
                if host:
                    chosenHost = host[0]
            else:
                chosenHost = host
        if chosenHost:
            #find host in df
            idx = hostDF['objID'] == chosenHost
            idx_transient = transientDF['Name'] == str(name)
            if hostDF.loc[idx, "TransientClass"].values[0] != "":
                print("Found a double!")
                hostDF = hostDF.append([hostDF[idx]], ignore_index=True)
                idx = hostDF.index[-1]
            hostDF.loc[idx, "TransientClass"] = transientDF.loc[idx_transient, 'Obj. Type'].to_string(index=False).strip()
            hostDF.loc[idx, "TransientName"] = transientDF.loc[idx_transient, 'Name'].to_string(index=False).strip()
            transCoord = SkyCoord(transientDF.loc[idx_transient, 'RA'], transientDF.loc[idx_transient, 'DEC'], unit=(u.deg, u.deg))
            if len(transCoord) > 1:
                transCoord = transCoord[0]
            hostDF.loc[idx, "TransientRA"] = transCoord.ra.deg
            hostDF.loc[idx, "TransientDEC"] = transCoord.dec.deg
            #adding all the extra columns that we haven't added yet
            for val in colNames:
                hostDF.loc[idx, "Transient"+val.replace(" ", "")] = transientDF.loc[idx_transient, val].to_string(index=False).strip()
    hostDF = hostDF[hostDF["TransientClass"] != ""]
    hostDF = hostDF.reset_index(drop=True)
    return hostDF
